@@ -3,12 +3,12 @@
 # @package      hubzero-expire-sessions
 # @file         expire-sessions
 # @author       Rick Kennell <kennell@purdue.edu>
+# @author       Pascal Meunier <pmeunier@purdue.edu>
 # @author       Nicholas J. Kisseberth <nkissebe@purdue.edu>
-# @author       Pascal Meunier 
-# @copyright    Copyright (c) 2005-2012 HUBzero Foundation, LLC.
+# @copyright    Copyright (c) 2005-2013 HUBzero Foundation, LLC.
 # @license      http://www.gnu.org/licenses/lgpl-3.0.html LGPLv3
 #
-# Copyright (c) 2005-2012 HUBzero Foundation, LLC.
+# Copyright (c) 2005-2013 HUBzero Foundation, LLC.
 #
 # This file is part of: The HUBzero(R) Platform for Scientific Collaboration
 #
@@ -28,46 +28,49 @@
 #
 # HUBzero is a registered trademark of HUBzero Foundation, LLC.s
 #
-import os,pwd,sys,time,MySQLdb,signal
+import os, pwd, sys, time, MySQLdb, signal
 
 os.chdir("/")
 
 foreground = 0
+DEBUG = 0
 
-expire_interval = 15
+expire_interval = 300
 logfile_name = "/var/log/expire-sessions/expire-sessions.log"
 logfile = None
 pidfile = "/var/run/expire-sessions.pid"
 
 if os.path.exists("/usr/lib/mw/bin/maxwell"):
-  mwprog="/usr/lib/mw/bin/maxwell"
+  mwprog = "/usr/lib/mw/bin/maxwell"
 elif os.path.exists("/usr/lib/hubzero/bin/maxwell"):
-  mwprog="/usr/lib/hubzero/bin/maxwell"
+  mwprog = "/usr/lib/hubzero/bin/maxwell"
 elif os.path.exists("/usr/bin/maxwell"):
-  mwprog="/usr/bin/maxwell"
+  mwprog = "/usr/bin/maxwell"
 else:
-  mwprog=""
+  mwprog = ""
 
 if os.path.exists("/etc/mw-client/mw-client.conf"):
-  config_file="/etc/mw-client/mw-client.conf"
+  config_file = "/etc/mw-client/mw-client.conf"
 elif os.path.exists("/etc/mw-www/maxwell.conf"):
-  config_file="/etc/mw-www/maxwell.conf"
+  config_file = "/etc/mw-www/maxwell.conf"
 elif os.path.exists("/etc/hubzero/maxwell.conf"):
-  config_file="/etc/hubzero/maxwell.conf"
+  config_file = "/etc/hubzero/maxwell.conf"
 elif os.path.exists("/etc/mw/maxwell.conf"):
-  config_file="/etc/mw/maxwell.conf"
+  config_file = "/etc/mw/maxwell.conf"
 else:
-  config_file=""
+  config_file = ""
 
 #==============================================================================
 # Default values.
 #==============================================================================
 
 mysql_host = ""
-mysql_user=""
-mysql_password=""
-mysql_db=""
-mysql_connect_attempts=120
+mysql_user =""
+mysql_password = ""
+mysql_db = ""
+mysql_connect_attempts = 120
+
+ANONYMOUS_SESSIONS = False
 
 #=============================================================================
 # Load the configuration and override the variables above.
@@ -104,18 +107,19 @@ def openlog(filename):
 # Log a message.
 #=============================================================================
 def log(msg):
-  if not foreground:
-    timestamp = "[" + time.asctime() + "] "
+  if foreground:
+    print msg
   else:
-    timestamp = ""
-  logfile.write(timestamp + msg + "\n")
-  logfile.flush()
+    timestamp = "[" + time.asctime() + "] "
+    #timestamp = ""
+    logfile.write(timestamp + msg + "\n")
+    logfile.flush()
 
 #=============================================================================
 # Create database connection.
 #=============================================================================
 def db_connect():
-  for x in range(0,mysql_connect_attempts):
+  for x in range(0, mysql_connect_attempts):
     try:
       db = MySQLdb.connect(host=mysql_host, user=mysql_user, passwd=mysql_password, db=mysql_db)
       #log("db_connect finished on iteration %d" % x)
@@ -127,14 +131,14 @@ def db_connect():
 #=============================================================================
 # MySQL helpers
 #=============================================================================
-def mysql(c,cmd):
+def mysql(c, cmd):
   try:
     count = c.execute(cmd)
     if c._warnings > 0:
       log("MySQL warning")
       log("SQL was: %s" % cmd)
     return c.fetchall()
-  except MySQLdb.MySQLError, (num, expl):
+  except MySQLdb.MySQLError, (count, expl):
     log("%s" % expl)
     log("SQL was: %s" % cmd)
     return ()
@@ -143,19 +147,16 @@ def mysql(c,cmd):
     log("SQL was: %s" % cmd)
     return ()
 
-def mysql_act(c,cmd):
+def mysql_act(c, cmd):
   try:
     count = c.execute(cmd)
     return ""
-  except MySQLdb.MySQLError, (num, expl):
+  except MySQLdb.MySQLError, (count, expl):
     return expl
 
 
 #=============================================================================
 # Check pidfile
-#
-# @TODO should probably check if process is actually an instance
-#       of expire-sessions
 #=============================================================================
 def checkpidfile():
 
@@ -164,19 +165,19 @@ def checkpidfile():
   except:
     return True
 
-  try:    
+  try:
     pid = int(f.read())
   except:
     #log("Unable to open pidfile %s" % pidfile)
     return False
-      
+
   if (pid <= 1):
     #log("Invalid pid found in pidfile %s" % pidfile)
     return False
 
   if os.path.exists("/proc/%s" % pid):
-  	print(" [expire-sessions (pid %s) already running] " % pid),
-  	return 0
+    print(" [expire-sessions (pid %s) already running] " % pid),
+    return 0
   else:
     return 1
 
@@ -190,7 +191,7 @@ def rmpidfile():
     #log("Unable to remove pidfile %s" % pidfile)
     return False
 
-  return true
+  return True
 
 #=============================================================================
 # Daemonize.
@@ -213,19 +214,19 @@ def daemonize():
 #=============================================================================
 # Change to different user.
 #=============================================================================
-def change_user(username):
+def change_www():
   if os.getuid() == 0:
     try:
-      _,_,uid,gid,_,_,_ = pwd.getpwnam("www-data")
-      os.setregid(gid,gid)
-      os.setreuid(uid,uid)
-    except:
+      _, _, uid, gid, _, _, _ = pwd.getpwnam("www-data")
+      os.setregid(gid, gid)
+      os.setreuid(uid, uid)
+    except OSError:
       pass
 
 #=============================================================================
 # Handle signals.
 #=============================================================================
-def sighandler(sig,frame):
+def sighandler(sig, frame):
   log("Caught signal %d.  Exiting." % sig)
   rmpidfile()
   sys.exit(2)
@@ -233,9 +234,9 @@ def sighandler(sig,frame):
 #=============================================================================
 # Do the work of terminating a session.
 #=============================================================================
-def terminate(sessnum,username,appname):
-  log("Terminating %d (%s,%s)" % (sessnum,username,appname))
-  status=os.system("%s stop reason=timeout sessnum=%d" % (mwprog,sessnum))
+def terminate(sessnum, username, appname):
+  log("Terminating %d (%s,%s)" % (sessnum, username, appname))
+  status = os.system("%s stop reason=timeout sessnum=%d" % (mwprog, sessnum))
   if status != 0:
     log("Error terminating session: %d" % status)
 
@@ -247,13 +248,14 @@ signal.signal(signal.SIGHUP, sighandler)
 signal.signal(signal.SIGINT, sighandler)
 signal.signal(signal.SIGQUIT, sighandler)
 signal.signal(signal.SIGTERM, sighandler)
- 
+
 if not checkpidfile():
+  print "checkpidfile failed"
   sys.exit(1)
 
 if not foreground:
   daemonize()
-  change_user("www-data")
+  change_www()
   openlog(logfile_name)
 
   log("Server is ready.")
@@ -261,37 +263,96 @@ if not foreground:
 # Delete old views without heartbeat
 db = db_connect()
 c = db.cursor()
-mysql(c, """delete from view where heartbeat <  DATE_ADD(CURDATE(),INTERVAL -7 DAY)""")
-db.close()
 
 while 1:
-  db = db_connect()
-  c = db.cursor()
-  arr = mysql(c, """
-           SELECT session.sessnum,session.username,appname,
-           timeout-TIMESTAMPDIFF(SECOND,accesstime,NOW()) AS remaining
-           FROM session
-           LEFT JOIN view ON session.sessnum = view.sessnum
-           LEFT JOIN job ON session.sessnum = job.sessnum
-           WHERE viewid IS NULL AND jobid IS NULL AND
-           timeout-TIMESTAMPDIFF(SECOND,accesstime,NOW()) < %d
-           ORDER BY REMAINING""" % expire_interval)
-  db.close()
-
   # Recheck, at the longest, every "expire_interval" seconds.
-  shortest = expire_interval
-  skew=0
-  for sessnum,username,appname,remaining in arr:
-    if foreground:
-      log("%d of %s to expire in %d seconds" % (sessnum,username,remaining))
-    if remaining < 1 + skew:
-      terminate(sessnum,username,appname)
+  sleeptime = expire_interval
+  # delete any orphaned views, which shouldn't exist unless there was a restart or crash
+  mysql(c, """DELETE FROM view WHERE heartbeat < DATE_ADD(NOW(), INTERVAL -1 DAY)""")
+  
+  # Anonymous sessions
+  if ANONYMOUS_SESSIONS:
+    anon_arr = mysql(c, """
+             SELECT session.sessnum, appname
+             FROM session LEFT JOIN view ON session.sessnum = view.sessnum
+             WHERE session.username = 'anonymous' 
+             AND viewid IS NULL""")
+    for sessnum, appname in anon_arr:
+      session = int(sessnum)
+      terminate(sessnum, 'anonymous', appname)
+      time.sleep(1) # Do not kill things off too quickly.
+      sleeptime -= 1
+  
+  # Regular sessions
+  # Find sessions that have been accessed more than "timeout" seconds ago
+  # and that have no views, and that could expire during our next sleep (expire_interval)
+  # skew: keep track of sleep time between each expired session
+  skew = 0 
+  arr = mysql(c, """
+    SELECT session.sessnum, session.username, appname,
+    timeout - TIMESTAMPDIFF(SECOND, accesstime, NOW()) AS remaining
+    FROM session 
+    LEFT JOIN view ON session.sessnum = view.sessnum
+    WHERE viewid IS NULL AND 
+    timeout - TIMESTAMPDIFF(SECOND, accesstime, NOW()) < %d
+    ORDER BY remaining""" % (expire_interval))
+
+  # Check if those sessions have jobs running
+  # there can be many jobs per session.  Expire session only if *all* jobs have had no heartbeat.
+  # Complication:  submit --local jobs that run as long as the session is open
+  # "submit --local" will have no heartbeat, whereas all other submit jobs will
+  # To handle this, expire session only if all job heartbeats are old.
+  for sessnum, username, appname, session_remaining in arr:
+    # Find the greatest remaining time for this session to stay alive
+    max_remaining = int(session_remaining)
+    # heartbeat is considered "stale" after 24 hours
+    # remaining = heartbeat - NOW() + 86400
+    job_arr = mysql(c, """SELECT 86400-TIMESTAMPDIFF(SECOND, heartbeat, NOW()) AS remaining
+      FROM job WHERE sessnum = "%s"
+      AND 86400-TIMESTAMPDIFF(SECOND, heartbeat, NOW()) > 0""" % sessnum)
+    if job_arr:
+      # Find the greatest reason this session should stay alive
+      for row in job_arr:
+        remaining = row[0]
+        if int(remaining)  > max_remaining:
+          max_remaining = int(remaining)
+
+    # How long should someone have to check the results of a simulation?  
+    # people have 24 hours to check results after the end of the last job.  
+    # If a job started at 9AM took 12 hours to run then a user would have until 9PM 
+    # the next day to check results.  It's always possible to check the results before 
+    # the session ends. 
+    # if any joblog entry is for a job that ended less than a day ago, pass
+    # remaining = start + walltime - NOW() + 86400
+    job_arr = mysql(c, """SELECT 86400 + walltime - TIMESTAMPDIFF(SECOND, start, NOW()) AS remaining
+      FROM joblog WHERE sessnum = "%s"
+      AND 86400 + walltime - TIMESTAMPDIFF(SECOND, start, NOW()) > 0""" % sessnum)
+    if job_arr:
+      for row in job_arr:
+        remaining = row[0]
+        if int(remaining)  > max_remaining:
+          max_remaining = int(remaining)
+      # log("%d of %s has a recently ended job %s" % (sessnum, username, job_arr))
+
+    if DEBUG:
+      log("%d of %s to expire in %d seconds" % (sessnum, username, max_remaining))
+    if max_remaining < 1 + skew:
+      terminate(sessnum, username, appname)
       time.sleep(1) # Do not kill things off too quickly.
       skew += 1
-    elif remaining < shortest:
-      shortest = remaining
-
-  shortest -= skew
-  if (shortest > 0):
-    time.sleep(shortest)
+    elif max_remaining < sleeptime:
+      sleeptime = max_remaining
+    if False:
+      if max_remaining > session_remaining:
+        # fake access as an optimization to avoid processing the same data again
+        # could this affect statistics?  Alternative is to check less often. ( 5 min instead of 15 seconds)
+        mysql(c, """UPDATE session SET accesstime = DATE_ADD(NOW(), INTERVAL %d - timeout SECOND)
+        WHERE sessnum = %d""" % (max_remaining - skew, session))
+      
 
+  sleeptime -= skew
+  if (sleeptime > 0):
+    db.close()
+    time.sleep(sleeptime)
+    db = db_connect()
+    c = db.cursor()
